// Copyright (C) 2001 Helmut Fahrion
// This file is released under the GPL.


// options
#include <getopt.h>
#include <stdio.h>
#include <sched.h>
#include <errno.h>
#include <sys/mman.h>		// mlockall
#include <unistd.h>		// execvp

char *programm;			// Fr Ausgaben
char version[] = "0.0.2\n";	// Programmversion
char *execstr;

struct timespec t;
__pid_t pid;
int policy;
struct sched_param param;

void
schedminmax ()
{
  printf ("sched_get_priority_min:\n");
  printf ("SCHED_OTHER = %d\n", sched_get_priority_min (SCHED_OTHER));
  printf ("SCHED_RR    = %d\n", sched_get_priority_min (SCHED_RR));
  printf ("SCHED_FIFO  = %d\n", sched_get_priority_min (SCHED_FIFO));

  printf ("sched_get_priority_max:\n");
  printf ("SCHED_OTHER = %d\n", sched_get_priority_max (SCHED_OTHER));
  printf ("SCHED_RR    = %d\n", sched_get_priority_max (SCHED_RR));
  printf ("SCHED_FIFO  = %d\n", sched_get_priority_max (SCHED_FIFO));

  // this is equal to sched_set_priority
  //printf("yield: %d\n", sched_yield());
}

char *
sched_strerror (int err)
{
  char *policystr;

  fprintf (stderr, "Error: ");
  switch (err)
    {
    case ESRCH:
      fprintf (stderr, "The process whose ID is pid=%d could not be found.\n",
	       pid);
      break;
    case EPERM:
      switch (policy)
	{
	case SCHED_FIFO:
	  policystr = "FIFO";
	  break;
	case SCHED_RR:
	  policystr = "RR";
	  break;
	case SCHED_OTHER:
	  policystr = "OTHER";
	  break;
	default:
	  policystr = "Unknown!";
	}
      fprintf (stderr,
	       "The calling process does not have appropriate privileges. Only root processes are allowed to activate the %s policies. The process calling sched_setscheduler needs an effective uid equal to the euid or uid of the process identified by pid=%d, or it must be a superuser process.\n",
	       policystr, pid);
      break;
    case EINVAL:
      fprintf (stderr,
	       "The scheduling policy is not one of the  recognized policies.\n");
      fprintf (stderr, " use:  1 for FIFO, 2 for RR and 0 for OTHER\n");
      break;
    }
}

int
set ()
{
  if (sched_setscheduler (pid, policy, &param))
    {
      strerror (errno);
      sched_strerror (errno);
      return 1;
    }
  return 0;
}

int
get ()
{
  policy = sched_getscheduler (pid);
  if (sched_getparam (pid, &param))
    {
      sched_strerror (errno);
      return 1;
    }
  if (sched_rr_get_interval (pid, &t))
    {
      sched_strerror (errno);
      return 1;
    }
  return 0;
}

void
printout ()
{
  printf ("pid: %d\tpriority: %d\tpolicy: ", pid, param.sched_priority);
  switch (policy)
    {
    case SCHED_FIFO:
      printf ("FIFO\t");
      break;
    case SCHED_RR:
      printf ("RR\t");
      break;
    case SCHED_OTHER:
      printf ("OTHER\t");
      break;
    default:
      printf ("Unknown! ");
    }

  printf ("timespec(sec,nsec): %d, %d\n", t.tv_sec, t.tv_nsec);
}

void
goexec (char *pgr, char *argv[], int pos)
{
  char **cmdarg;
  int x;

  cmdarg = argv + pos;

  if (mlockall (MCL_FUTURE))
    {
      fprintf (stderr, "%s\n", strerror (errno));
      exit (-1);
    }
  else
    {
      // execute programm
      set ();
      printout ();

      printf ("execvp: ");

      for (x = 0; cmdarg[x]; x++)
	{
	  printf ("%s ", cmdarg[x]);
	}
      printf ("\n");
      fflush (stdout);

      execvp (pgr, cmdarg);
    }
}


void
usage ()
{
  printf
    ("Usage: %s  [--list|-l ] [--help|--hilfe|-h|-?] [--version|-v] [--priority|-p num] [ -polcy|-p polycinum ] [[--set|-s ] | [--get|-g] ] pid | [ --exec|-e  programm param1..paramX ] \n",
     programm);
  printf ("Commands:\n");
  printf (" --set -s:  set the data\n");
  printf (" --get -g:  get the data\n");
  printf (" --list -l:  print out avilable scheduling alogrithmus\n");
  printf (" --exec -e:  set and exec command with mmapall function\n");
  printf (" --help --hilfe -h -?:  print out help\n");
  printf (" --version -v:  print out this program version\n");

  printf ("\nParameters:\n");
  printf (" --priority -p:  the priority\n");
  printf (" --policy -a:  the method\n");

  printf (" use:  1 for FIFO, 2 for RR and 0 for OTHER\n");
  printf (" use --exec parameter folowed by parameters as last arguments\n");
  printf
    ("   schedule --priority 10 --policy 1 --exec wwwoffled -c /etc/wwwoffle/wwwoffle.conf\n");
}

int
main (int argc, char *argv[])
{
  int opterr, opt, cmd, x;

  enum
  {
    NOP,
    GET,
    SET,
    TIME,
  };

  static const struct option long_options[] = {
    {"list", 0, 0, 'l'},
    {"hilfe", 0, 0, '?'},
    {"help", 0, 0, 'h'},
    {"get", 1, 0, 'g'},
    {"set", 1, 0, 's'},
    {"exec", 1, 0, 'e'},

    {"version", 0, 0, 'v'},

    {"priority", 1, 0, 'p'},
    {"policy", 1, 0, 'a'},

    {0, 0, 0, 0}
  };

  programm = argv[0];
  opterr = policy = param.sched_priority = x = 0;
  cmd = NOP;

  if (argc < 2)
    {
      usage ();
      return 1;
    }

  while ((opt =
	  getopt_long (argc, argv, "+h?vltg:s:a:p:e:m:", long_options,
		       NULL)) != EOF)
    {
      x++;
      switch (opt)
	{
	case 'e':
	  x++;
	  goexec (optarg, argv, x);
	  break;

	case 'p':
	  x++;
	  param.sched_priority = atol (optarg);
	  break;

	case 'a':
	  x++;
	  policy = atoi (optarg);
	  break;

	case 's':
	  pid = atol (optarg);
	  cmd = SET;
	  break;

	case 'g':
	  pid = atol (optarg);
	  cmd = GET;
	  break;

	case 'l':
	  schedminmax ();
	  break;

	case 'h':
	case '?':
	  usage ();
	  return 0;

	case 'v':
	  //printf ("Linux %s version %s\n", programm, version);
	  printf (version);
	  return 0;

	default:
	  printf ("%s: Ambiguous or unknown options!\n", programm);
	  usage ();
	  return 1;
	}
    }

  if (cmd)
    {
      if (cmd == SET)
	{
	  if (set ())
	    return 1;
	}

      if (cmd == GET)
	{
	  if (get ())
	    return 1;
	}

      printout ();
    }

  return 0;
}
